/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.organdonor.personinformation;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.core.type.TypeReference;
import com.fasterxml.jackson.databind.ObjectMapper;
import dk.sundhedsdatastyrelsen.organdonor.backgroundjob.health.StatusProvider;
import org.apache.commons.codec.Charsets;
import org.apache.http.Header;
import org.apache.http.HttpResponse;
import org.apache.http.HttpStatus;
import org.apache.http.client.methods.CloseableHttpResponse;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.client.methods.HttpUriRequest;
import org.apache.http.client.utils.URIBuilder;
import org.apache.http.entity.StringEntity;
import org.apache.http.impl.client.CloseableHttpClient;
import org.apache.log4j.Logger;
import org.openapitools.model.Age;
import org.openapitools.model.Custody;
import org.openapitools.model.StatusCode;

import java.io.IOException;
import java.net.URISyntaxException;
import java.time.Duration;
import java.time.Instant;
import java.time.ZoneId;
import java.time.format.DateTimeFormatter;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.function.Function;
import java.util.stream.Collectors;

public class PersonInformationClient implements StatusProvider {

    private static final String APPLICATION_JSON = "application/json";

    private static final String CONTENT_TYPE = "Content-Type";

    private static final Logger LOGGER = Logger.getLogger(PersonInformationClient.class);

    private final ObjectMapper objectMapper = new ObjectMapper();

    private final CloseableHttpClient httpClient;

    private final ServiceErrorCallback serviceErrorCallback;
    private final int errorTolerance;

    private final String url;

    public PersonInformationClient(CloseableHttpClient httpClient, String url, ServiceErrorCallback serviceErrorCallback, int errorTolerance) {
        this.httpClient = httpClient;
        this.url = url;
        this.serviceErrorCallback = serviceErrorCallback;
        this.errorTolerance = errorTolerance;
    }
    public PersonInformationClient(CloseableHttpClient httpClient, String url, Duration errorCountDuration, ZoneId zoneId, int errorTolerance) {
        this(httpClient, url, new ServiceErrorCallback(errorCountDuration, zoneId), errorTolerance);
    }

    public Custody getCustody(String patientId) {
        HttpUriRequest cprRequest = buildGetCustodyRequest(patientId);
        Map<String, Object> response = invokeService(cprRequest, this::parseResponseAsJson);
        return parseCustodyFromResponse(response);
    }

    public String getName(String patientId) {
        HttpUriRequest cprRequest = buildGetNameRequest(patientId);
        Map<String, Object> response = invokeService(cprRequest, this::parseResponseAsJson);
        return parseNameFromResponse(response);
    }

    public int getAge(String patientId) {
        HttpUriRequest cprRequest = buildGetAgeRequest(patientId);
        Map<String, Object> response = invokeService(cprRequest, this::parseResponseAsJson);
        return parseAgeFromResponse(response);
    }

    public List<String> getDeceased(List<String> cprList, Instant deceasedDate) {
        HttpUriRequest cprRequest = buildGetDeceasedRequest(cprList, deceasedDate);
        List<Map<String, Object>> response = invokeService(cprRequest, this::parseJsonResponseAsList);
        return parseDeceasedFromResponse(response);
    }

    public String getStatus(String patientId) {
        HttpUriRequest cprRequest = buildGetStatusRequest(patientId);
        Map<String, Object> response = invokeService(cprRequest, this::parseResponseAsJson);
        return parseStausCodeFromResponse(response);
    }

    public List<String> doFilterByStatusCode(List<String> cprList, Instant statusCodeSinceDate, String statusCode) {
        HttpUriRequest cprRequest = buildPostFilterByStatusCode(cprList, statusCodeSinceDate, statusCode);
        List<Map<String, Object>> response = invokeService(cprRequest, this::parseJsonResponseAsList);
        return parseStatusCodesFromResponse(response);
    }

    private HttpUriRequest buildGetCustodyRequest(String cpr) {
        // Example request string: http://localhost:8086/stamdata-personinformation/person/1212121212/custody
        return new HttpGet(String.format("%s/person/%s/custody", url, cpr));
    }

    private HttpUriRequest buildGetNameRequest(String cpr) {
        // Example request string: http://localhost:8086/stamdata-personinformation/person/1212121212/name
        return new HttpGet(String.format("%s/person/%s/name", url, cpr));
    }

    private HttpUriRequest buildGetAgeRequest(String cpr) {
        // Example request string: http://localhost:8086/stamdata-personinformation/person/1212121212/age
        return new HttpGet(String.format("%s/person/%s/age", url, cpr));
    }

    private HttpUriRequest buildGetDeceasedRequest(List<String> cprList, Instant deceasedSince) throws PersonInformationException {
        // Example request string: http://stamdatahost/stamdata-personinformation/person/deceased?cpr=0304784567&cpr=1508902650&dato=2023-07-21T17:32:28
        String utcDateString = DateTimeFormatter.ISO_INSTANT.format(deceasedSince);
        String urlWithContext = String.format("%s/person/deceased?", url);
        try {
            URIBuilder builder = new URIBuilder(urlWithContext);
            for (String cpr : cprList) {
                builder.addParameter("cpr", cpr);
            }
            builder.addParameter("dato", utcDateString);
            return new HttpGet(builder.build());
        } catch (URISyntaxException e) {
            String error = ErrorMessage.PERSON_INFORMATION_SERVICE_CALL_FAILED.getMessage();
            LOGGER.error(error);
            throw new PersonInformationException(error, e);
        }
    }

    private HttpUriRequest buildGetStatusRequest(String cpr) {
        // Example request string: http://localhost:8086/stamdata-personinformation/person/1212121212/status
        return new HttpGet(String.format("%s/person/%s/status", url, cpr));
    }

    private HttpUriRequest buildPostFilterByStatusCode(List<String> cprs, Instant statusCodeSince, String statusCode) {
        // Example request string: http://localhost:8086/stamdata-personinformation/2024/08/01/person/filterByStatusCode?statusCode=?&dato=?
        String utcDateString = DateTimeFormatter.ISO_INSTANT.format(statusCodeSince);
        String urlWithContext = String.format("%s/person/filterByStatusCode?", url);
        try {
            URIBuilder builder = new URIBuilder(urlWithContext);
            builder.addParameter("statusCode", statusCode);
            builder.addParameter("dato", utcDateString);
            HttpPost post = new HttpPost(builder.build());
            post.setHeader(CONTENT_TYPE, APPLICATION_JSON);
            String json = new ObjectMapper().writeValueAsString(cprs);
            StringEntity data = new StringEntity(json, Charsets.ISO_8859_1);
            post.setEntity(data);
            return post;
        } catch (URISyntaxException | JsonProcessingException e) {
            String error = ErrorMessage.PERSON_INFORMATION_SERVICE_CALL_FAILED.getMessage();
            LOGGER.error(error);
            throw new PersonInformationException(error, e);
        }
    }

    private <E> E invokeService(HttpUriRequest request, Function<HttpResponse, E> responseParser) {
        try(CloseableHttpResponse response = httpClient.execute(request)) {
            E result = handleResponse(response, responseParser);
            serviceErrorCallback.registerServiceSuccess();
            return result;
        } catch (IOException e) {
            String errorMessage = ErrorMessage.MAX_ALLOWED_SIZE_ERROR.getMessage();
            LOGGER.error(errorMessage, e);
            serviceErrorCallback.registerServiceError();
            throw new PersonInformationException(errorMessage, e);
        } catch (PersonInformationException e) {
            serviceErrorCallback.registerServiceError();
            throw e;
        }
    }

    private <E> E handleResponse(HttpResponse response, Function<HttpResponse, E> responseParser) {
        int statusCode = response.getStatusLine().getStatusCode();
        switch (statusCode) {
            case HttpStatus.SC_OK: {
                boolean contentTypeOk = contentTypeOk(response);
                //Map<String, Object> responseData = parseResponseAsJson(response);
                E responseData = responseParser.apply(response);
                if (!contentTypeOk) {
                    String error = ErrorMessage.PERSON_INFORMATION_SERVICE_CALL_CONTENT_TYPE_WRONG.getMessage();
                    LOGGER.error(error);
                    throw new PersonInformationException(error);
                }
                return responseData;
            }
            case HttpStatus.SC_NOT_FOUND: {
                boolean contentTypeOk = contentTypeOk(response);
                if (!contentTypeOk) {
                    String error = ErrorMessage.PERSON_INFORMATION_SERVICE_CALLERROR404.getMessage();
                    throw new PersonInformationException(error);
                }
                return responseParser.apply(response);
            }
            default:
                throw new PersonInformationException(ErrorMessage.PERSON_INFORMATION_SERVICE_CALL_UNKNOWN_STATUS.getAndFormatMessage(Integer.toString(statusCode)));
        }
    }


    private Custody parseCustodyFromResponse(Map<String, Object> response) {
        if (response.isEmpty()) {
            return null;
        } else {
            return objectMapper.convertValue(response, Custody.class);
        }
    }

    private String parseNameFromResponse(Map<String, Object> response) {
        if (response.isEmpty()) {
            return "";
        } else {
            String firstName = (response.get("givenName") == null ? "" : response.get("givenName").toString());
            String middleName = (response.get("middleName") == null || response.get("middleName").toString().isEmpty() ? " " : " " + response.get("middleName").toString() + " ");
            String lastName = (response.get("familyName") == null ? "" : response.get("familyName").toString());

            return firstName + middleName + lastName;
        }
    }

    private int parseAgeFromResponse(Map<String, Object> response) {
        if (response.isEmpty()) {
            return 0;
        } else {
            Age age = objectMapper.convertValue(response, Age.class);
            return (age != null) ? age.getAge() : 0;
       }
    }

    private List<String> parseDeceasedFromResponse(List<Map<String, Object>> response) {
        if (response.isEmpty()) {
            return Collections.emptyList();
        } else {
            return response.stream()
                    .map(map -> map.get("cpr").toString())
                    .collect(Collectors.toList());
        }
    }

    private String parseStausCodeFromResponse(Map<String, Object> response) {
        if (response.isEmpty()) {
            return null;
        } else {
            StatusCode statusCode = objectMapper.convertValue(response, StatusCode.class);
            return (statusCode != null) ? statusCode.getStatusCode() : null;
        }
    }

    private List<String> parseStatusCodesFromResponse(List<Map<String, Object>> response) {
        if (response.isEmpty()) {
            return Collections.emptyList();
        } else {
            return response.stream()
                    .map(map -> map.get("cpr").toString())
                    .collect(Collectors.toList());
        }
    }

    private Map<String, Object> parseResponseAsJson(HttpResponse response) {
        try {
            return objectMapper.readValue(response.getEntity().getContent(), Map.class);
        } catch (UnsupportedOperationException | IOException e) {
            LOGGER.error("Kunne ikke parse response body til json", e);
            throw new PersonInformationException(ErrorMessage.PERSON_INFORMATION_SERVICE_CALL_PARSE_ERROR.getMessage(), e);
        }
    }

    private List<Map<String, Object>> parseJsonResponseAsList(HttpResponse response) throws PersonInformationException {
        try {
            return objectMapper.readValue(response.getEntity().getContent(), new TypeReference<List<Map<String, Object>>>(){});
        } catch (UnsupportedOperationException | IOException e) {
            LOGGER.error("Kunne ikke parse json til List<Map<String, Object>> type", e);
            throw new PersonInformationException("Fejl i fortolkning af svar fra PersonInformation", e);
        }
    }

    private boolean contentTypeOk(HttpResponse response) {
        Header headers = response.getFirstHeader(CONTENT_TYPE);
        return headers != null && headers.getValue().equalsIgnoreCase(APPLICATION_JSON);
    }

    @Override
    public boolean isOk() {
        return serviceErrorCallback.getErrorCount() <= errorTolerance;
    }
}
