/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.organdonor.cleanup.factory;

import dk.sundhedsdatastyrelsen.organdonor.backgroundjob.stack.StackOperation;
import dk.sundhedsdatastyrelsen.organdonor.backgroundjob.util.DateTimeProvider;
import dk.sundhedsdatastyrelsen.organdonor.cleanup.config.CleanupProperties;
import dk.sundhedsdatastyrelsen.organdonor.cleanup.service.CleanupService;
import dk.sundhedsdatastyrelsen.organdonor.cleanup.stack.deceased.CPRBatchDeceasedCleanupSupplier;
import dk.sundhedsdatastyrelsen.organdonor.cleanup.stack.emigrated.CPRBatchEmigratedCleanupSupplier;
import dk.sundhedsdatastyrelsen.organdonor.cleanup.stack.deceased.CPRDeceasedCleanupOperation;
import dk.sundhedsdatastyrelsen.organdonor.cleanup.stack.deceased.CPRPrefixDeceasedCleanupSupplier;
import dk.sundhedsdatastyrelsen.organdonor.cleanup.stack.emigrated.CPRPrefixEmigratedCleanupSupplier;
import dk.sundhedsdatastyrelsen.organdonor.cleanup.stack.emigrated.CPREmigratedCleanupOperation;
import dk.sundhedsdatastyrelsen.organdonor.service.PersonInformationService;

import java.util.List;

public class CleanupOperationsFactory {

    private final CleanupService cleanupService;
    private final PersonInformationService personInformationService;
    private final CleanupProperties cleanupProperties;
    private final DateTimeProvider dateTimeProvider;

    public CleanupOperationsFactory(CleanupProperties cleanupProperties, CleanupService cleanupService, PersonInformationService personInformationService, DateTimeProvider dateTimeProvider) {
        this.cleanupService = cleanupService;
        this.personInformationService = personInformationService;

        this.dateTimeProvider = dateTimeProvider;

        this.cleanupProperties = cleanupProperties;
    }

    /*
     * CPRPrefixDeceasedCleanupSupplier - Laver en liste med med CPRBatchDeceasedCleanupSupplier
     */
    public StackOperation createCPRPrefixDeceasedCleanupSupplier(String personIdPrefix) {
        return new CPRPrefixDeceasedCleanupSupplier(this, cleanupProperties, cleanupService, personIdPrefix);
    }

    /*
     * CPRBatchDeceasedCleanupSupplier - Laver en liste med CPRCleanupDeceasedOperation
     */
    public StackOperation createCPRBatchDeceasedCleanupSupplier(String personIdPrefix, List<String> records) {
        return new CPRBatchDeceasedCleanupSupplier(this, cleanupProperties, personInformationService, dateTimeProvider, personIdPrefix, records);
    }

    /*
     * CPRCleanupDeceasedOperation - Udfører sletning vha. PersonIdentifier
     */
    public StackOperation createCPRCleanupDeceasedOperation(String personIdPrefix, List<String> personIds) {
        return new CPRDeceasedCleanupOperation(cleanupService, dateTimeProvider, personIdPrefix, personIds);
    }

    /*
     * CPRPrefixEmigratedCleanupSupplier - Laver en liste med med CPRBatchEmigratedCleanupSupplier
     */
    public StackOperation createCPRPrefixEmigratedCleanupSupplier(String personIdPrefix) {
        return new CPRPrefixEmigratedCleanupSupplier(this, cleanupProperties, cleanupService, personIdPrefix);
    }

    /*
     * CPRBatchEmigratedCleanupSupplier - Laver en liste med CPRCleanupEmigratedOperation
     */
    public StackOperation createCPRBatchEmigratedCleanupSupplier(String personIdPrefix, List<String> records) {
        return new CPRBatchEmigratedCleanupSupplier(this, cleanupProperties, personInformationService, dateTimeProvider, personIdPrefix, records);
    }

    /*
     * CPRCleanupEmigratedOperation - Udfører opdatering af tildato vha. PersonIdentifier
     */
    public StackOperation createCPRCleanupEmigratedOperation(String personIdPrefix, List<String> personIds) {
        return new CPREmigratedCleanupOperation(cleanupService, dateTimeProvider, personIdPrefix, personIds);
    }
}
