/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.organdonor.cleanup.stack.emigrated;

import dk.sundhedsdatastyrelsen.organdonor.backgroundjob.stack.StackOperation;
import dk.sundhedsdatastyrelsen.organdonor.backgroundjob.util.BatchUtil;
import dk.sundhedsdatastyrelsen.organdonor.backgroundjob.util.DateTimeProvider;
import dk.sundhedsdatastyrelsen.organdonor.cleanup.config.CleanupProperties;
import dk.sundhedsdatastyrelsen.organdonor.cleanup.factory.CleanupOperationsFactory;
import dk.sundhedsdatastyrelsen.organdonor.cleanup.service.CleanupException;
import dk.sundhedsdatastyrelsen.organdonor.cleanup.service.ErrorMessages;
import dk.sundhedsdatastyrelsen.organdonor.cleanup.stack.AbstractCleanupSupplier;
import dk.sundhedsdatastyrelsen.organdonor.service.PersonInformationService;

import java.time.Instant;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.stream.Collectors;

public class CPRBatchEmigratedCleanupSupplier extends AbstractCleanupSupplier {
    private static final String CPR_STATUS_CODE_INACTIVE_EMIGRATED = "80";

    private final PersonInformationService personInformationService;
    private final DateTimeProvider dateTimeProvider;
    private final String prefix;
    private final List<String> cprList;
    private final int personinformationBatchSize;

    public CPRBatchEmigratedCleanupSupplier(CleanupOperationsFactory cleanupOperationsFactory, CleanupProperties cleanupProperties, PersonInformationService personInformationService, DateTimeProvider dateTimeProvider, String prefix, List<String> cprList) {
        super(cleanupOperationsFactory, cleanupProperties);
        this.personInformationService = personInformationService;
        this.dateTimeProvider = dateTimeProvider;
        this.prefix = prefix;
        this.cprList = cprList;
        this.personinformationBatchSize = cleanupProperties.getPersoninformationEmigratedBatchSize();
        log.debug("CPRBatchEmigratedCleanupSupplier created with cprList size: " + cprList.size() + ", personinformationBatchSize: " + personinformationBatchSize);
    }

    @Override
    public String getIdentity() {
        return getClass().getSimpleName()+"{prefix=" + prefix + ", size=" + cprList.size() + "}";
    }

    @Override
    public List<StackOperation> run() {
        log.debug("CPRBatchEmigratedCleanupSupplier executing");
        Instant emigratedTimeNow = dateTimeProvider.now().toInstant();
        /* Da REST kaldet til PersonInformation Emigrated er et HTTP GET kald, så kan vi kun sende en lille mængde cprnr. med som parametre ad gangen. */
        Collection<List<String>> batches = BatchUtil.createBatches(cprList, personinformationBatchSize);
        log.debug("Created " + batches.size() + " batches for cprList size: " + cprList.size());
        /* Hver delmængde af cprnr. tjekkes om det tilhører en udrejst. Alle de cprnr. der gør det, samles i én liste.*/
        List<String> EmigratedList = batches.stream().map(batch -> checkIfEmigrated(batch, emigratedTimeNow)).flatMap(Collection::stream).collect(Collectors.toList());
        log.debug("Found " + EmigratedList.size() + " Emigrated persons for cprList size: " + cprList.size());
        return EmigratedList.isEmpty() ? Collections.emptyList() : Collections.singletonList(cleanupOperationsFactory.createCPRCleanupEmigratedOperation(prefix, EmigratedList));
    }

    private List<String> checkIfEmigrated(List<String> cprList, Instant latestEmigratedTime) {
        try {
            return personInformationService.filterEmigrated(cprList, latestEmigratedTime, CPR_STATUS_CODE_INACTIVE_EMIGRATED);
        } catch (CleanupException e) {
            throw e;
        } catch (Exception e) { // catch all, fordi vi gerne vil have en god tekst til alarm endpointet.
            throw new CleanupException(ErrorMessages.cleanupRetrieveEmigrated(latestEmigratedTime), e);
        }
    }
}
