/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.organdonor.cleanup.stack.emigrated;

import dk.sundhedsdatastyrelsen.organdonor.backgroundjob.stack.StackOperation;
import dk.sundhedsdatastyrelsen.organdonor.backgroundjob.util.BatchUtil;
import dk.sundhedsdatastyrelsen.organdonor.cleanup.config.CleanupProperties;
import dk.sundhedsdatastyrelsen.organdonor.cleanup.factory.CleanupOperationsFactory;
import dk.sundhedsdatastyrelsen.organdonor.cleanup.service.CleanupException;
import dk.sundhedsdatastyrelsen.organdonor.cleanup.service.CleanupService;
import dk.sundhedsdatastyrelsen.organdonor.cleanup.service.ErrorMessages;
import dk.sundhedsdatastyrelsen.organdonor.cleanup.stack.AbstractCleanupSupplier;

import java.util.Collection;
import java.util.List;
import java.util.stream.Collectors;

public class CPRPrefixEmigratedCleanupSupplier extends AbstractCleanupSupplier {
    // Størrelsen på hvert batch af personer, der skal behandles
    private final int batchSize;
    // Prefixet for CPR-numre, der skal behandles
    private final String prefix;
    // Service til at hente personer, der skal behandles
    private final CleanupService cleanupService;

    public CPRPrefixEmigratedCleanupSupplier(CleanupOperationsFactory cleanupOperationsFactory, CleanupProperties cleanupProperties, CleanupService cleanupService, String prefix) {
        super(cleanupOperationsFactory);
        this.cleanupService = cleanupService;
        this.prefix = prefix;
        this.batchSize = cleanupProperties.getDeletionEmigratedBatchSize();
        log.debug("CPRPrefixEmigratedCleanupSupplier created with prefix: " + prefix + ", batchSize: " + batchSize);
    }

    @Override
    public String getIdentity() {
        return getClass().getSimpleName()+"{prefix=" + prefix + ", batchSize=" + batchSize + "}";
    }

    @Override
    public List<StackOperation> run() {
        log.debug("CPRPrefixEmigratedCleanupSupplier executing");
        // Hent en liste af personer, der skal behandles
        List<String> personIdsToProcess = findPersonIdsToProcess();
        log.debug("Found " + personIdsToProcess.size() + " persons to process for prefix: " + prefix);
        // Opret en liste af batches, der skal behandles
        Collection<List<String>> batches = BatchUtil.createShuffledBatches(personIdsToProcess, batchSize);
        log.debug("Created " + batches.size() + " batches for prefix: " + prefix);
        // Opret en liste af operationer, der skal udføres. Et for hvert batch
        return batches.stream()
                .map(batch -> cleanupOperationsFactory.createCPRBatchEmigratedCleanupSupplier(prefix, batch))
                .collect(Collectors.toList());
    }

    /**
     * Henter en liste af personer, der skal behandles
     * @return Liste af personer, der skal behandles
     */
    private List<String> findPersonIdsToProcess() {
        try {
            return cleanupService.getValidPersonIdBatchByPersonIdPrefix(prefix);
        } catch(CleanupException e){
            throw e;
        } catch (Exception e) {
            throw new CleanupException(ErrorMessages.cleanupErrorReading(prefix), e);
        }
    }
}
