/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.organdonor.cleanup.stack.emigrated;

import dk.sundhedsdatastyrelsen.organdonor.backgroundjob.stack.StackOperation;
import dk.sundhedsdatastyrelsen.organdonor.backgroundjob.util.DateTimeProvider;
import dk.sundhedsdatastyrelsen.organdonor.cleanup.config.CleanupProperties;
import dk.sundhedsdatastyrelsen.organdonor.cleanup.factory.CleanupOperationsFactory;
import dk.sundhedsdatastyrelsen.organdonor.cleanup.service.CleanupException;
import dk.sundhedsdatastyrelsen.organdonor.cleanup.service.CleanupService;
import dk.sundhedsdatastyrelsen.organdonor.service.PersonInformationService;
import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.junit.MockitoJUnitRunner;

import java.util.Collections;
import java.util.List;
import java.util.stream.Collectors;
import java.util.stream.Stream;

/**
 * The MIT License
 * <p>
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 * <p>
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 * <p>
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 * <p>
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 * <p>
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 * <p>
 * Created by MacAni on 26/06/2017 11.37.
 */

@RunWith(MockitoJUnitRunner.class)
public class CPRPrefixEmigratedCleanupSupplierTest {

    //CleanupOperationsFactory cleanupOperationsFactory, CleanupProperties cleanupProperties, CleanupService cleanupService, String prefix
    @Mock
    CleanupProperties cleanupProperties;
    @Mock
    CleanupService cleanupService;
    @Mock
    PersonInformationService personInformationService;
    @Mock
    DateTimeProvider dateTimeProvider;

    private static final String PREFIX = "0306";
    private static final List<String> PERSON_IDS = Stream.of("0306000000", "0306111111","0306222222","0306333333","0306444444","0306555555","0306666666","0306777777","0306888888","0306999999").collect(Collectors.toList());



    CleanupOperationsFactory cleanupOperationsFactory;

    CPRPrefixEmigratedCleanupSupplier cprPrefixEmigratedCleanupSupplier;

    @Before
    public void setUp(){
        Mockito.when(cleanupService.getValidPersonIdBatchByPersonIdPrefix(PREFIX)).thenReturn(PERSON_IDS);
        Mockito.when(cleanupProperties.getDeletionEmigratedBatchSize()).thenReturn(3);

        cleanupOperationsFactory = Mockito.spy(new CleanupOperationsFactory(cleanupProperties, cleanupService, personInformationService, dateTimeProvider));
        cprPrefixEmigratedCleanupSupplier = new CPRPrefixEmigratedCleanupSupplier(cleanupOperationsFactory, cleanupProperties, cleanupService, PREFIX);
    }

    @Test
    public void testCPRPrefixEmigratedCleanupSupplier10IdsBatchesOf3(){
        // List of 10 cpr numbers in PERSON_IDS. 3 is the batch size. So, 4 batches will be created.

        List<StackOperation> stackOperations = cprPrefixEmigratedCleanupSupplier.run();
        Assert.assertEquals(4, stackOperations.size());
        Mockito.verify(cleanupOperationsFactory, Mockito.times(4)).createCPRBatchEmigratedCleanupSupplier(Mockito.eq(PREFIX), Mockito.anyList());
    }

    @Test
    public void testCPRPrefixEmigratedCleanupSupplier0IdsBatchesOf3(){
        // List of 0 cpr numbers in PERSON_IDS. 3 is the batch size. So, 0 batches will be created.
        Mockito.when(cleanupService.getValidPersonIdBatchByPersonIdPrefix(PREFIX)).thenReturn(Collections.emptyList());

        List<StackOperation> stackOperations = cprPrefixEmigratedCleanupSupplier.run();
        Assert.assertEquals(0, stackOperations.size());
        Mockito.verify(cleanupOperationsFactory, Mockito.times(0)).createCPRBatchEmigratedCleanupSupplier(Mockito.eq(PREFIX), Mockito.anyList());
    }

    @Test
    public void testCPRPrefixEmigratedCleanupSupplier10IdsBatchesOf100(){
        // List of 10 cpr numbers in PERSON_IDS. 100 is the batch size. So, 1 batch will be created.
        Mockito.when(cleanupProperties.getDeletionEmigratedBatchSize()).thenReturn(100);
        cprPrefixEmigratedCleanupSupplier = new CPRPrefixEmigratedCleanupSupplier(cleanupOperationsFactory, cleanupProperties, cleanupService, PREFIX);

        List<StackOperation> stackOperations = cprPrefixEmigratedCleanupSupplier.run();
        Assert.assertEquals(1, stackOperations.size());
        Mockito.verify(cleanupOperationsFactory, Mockito.times(1)).createCPRBatchEmigratedCleanupSupplier(Mockito.eq(PREFIX), Mockito.anyList());
    }

    @Test
    public void testCPRPrefixEmigratedCleanupSupplierFailsFindingIdsToProcess(){
        // List of 10 cpr numbers in PERSON_IDS. 3 is the batch size. So, 4 batches will be created.
        Mockito.when(cleanupService.getValidPersonIdBatchByPersonIdPrefix(PREFIX)).thenThrow(new RuntimeException("Test Error"));

        CleanupException e = Assert.assertThrows(CleanupException.class, cprPrefixEmigratedCleanupSupplier::run);
        Assert.assertEquals("Der opstod en fejl i forbindelse med læsning af LogEntry til sletning med personId prefix: 0306", e.getMessage());
        Assert.assertEquals("Test Error", e.getCause().getMessage());
    }
}
