/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.organdonor.cleanup.ws;

import dk.sundhedsdatastyrelsen.organdonor.backgroundjob.util.DateTimeProvider;
import dk.sundhedsdatastyrelsen.organdonor.backgroundjob.ws.AbstractOperationsContextListener;
import dk.sundhedsdatastyrelsen.organdonor.backgroundjob.ws.AbstractOperationsServlet;
import dk.sundhedsdatastyrelsen.organdonor.cleanup.config.CleanupProperties;
import dk.sundhedsdatastyrelsen.organdonor.cleanup.config.CleanupTestConfig;
import dk.sundhedsdatastyrelsen.organdonor.dao.OrganDonorDAO;
import dk.sundhedsdatastyrelsen.organdonor.personinformation.PersonInformation;
import org.apache.log4j.ConsoleAppender;
import org.apache.log4j.Level;
import org.apache.log4j.LogManager;
import org.junit.Assert;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.MockitoAnnotations;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.config.AutowireCapableBeanFactory;
import org.springframework.context.annotation.AnnotationConfigApplicationContext;
import org.springframework.mock.web.MockHttpServletRequest;
import org.springframework.mock.web.MockHttpServletResponse;
import org.springframework.mock.web.MockServletConfig;
import org.springframework.test.annotation.DirtiesContext;
import org.springframework.test.context.ContextConfiguration;
import org.springframework.test.context.junit4.SpringRunner;

import javax.annotation.Resource;
import java.util.Arrays;
import java.util.List;

@RunWith(SpringRunner.class)
@ContextConfiguration(classes = {CleanupTestConfig.class})
@DirtiesContext(classMode = DirtiesContext.ClassMode.AFTER_EACH_TEST_METHOD)
public class OrganDonorDeceasedCleanupServletTest {

    // Mocking HTTP Servlet Objects
    MockServletConfig mockServletConfig = new MockServletConfig("OrganDonorDeceasedCleanupServlet");
    MockHttpServletRequest mockHttpServletRequest = new MockHttpServletRequest();
    MockHttpServletResponse mockHttpServletResponse = new MockHttpServletResponse();

    @Mock
    AnnotationConfigApplicationContext mockApplicationContext;
    @Mock
    AutowireCapableBeanFactory mockAutowireCapableBeanFactory;

    @Resource(name="mockedOrganDonorDAO")
    OrganDonorDAO organDonorDAO;

    @Autowired
    PersonInformation personInformation;

    @Autowired
    DateTimeProvider dateTimeProvider;

    @Autowired
    CleanupProperties cleanupProperties;

    @Autowired
    AbstractOperationsServlet organDonorDeceasedCleanupServlet;

    @BeforeClass
    public static void setUpClass() {
        // This method can be used to set up any static resources or configurations needed for the tests
        ConsoleAppender consoleAppender = new ConsoleAppender();
        consoleAppender.setLayout(new org.apache.log4j.PatternLayout("%d{ISO8601} %p [%t] %c: %m%n"));
        consoleAppender.setThreshold(Level.INFO);
        consoleAppender.activateOptions();
        LogManager.getRootLogger().setLevel(Level.INFO);
        LogManager.getRootLogger().addAppender(consoleAppender);
    }

    @Before
    public void setUp() throws Exception {
        // Arrange
        MockitoAnnotations.initMocks(this);
        // Initialize the servlet with the mock configuration
        mockServletConfig.getServletContext().setAttribute(AbstractOperationsContextListener.ANNOTATION_CONFIG_CONTEXT, mockApplicationContext);
        Mockito.doReturn(mockAutowireCapableBeanFactory).when(mockApplicationContext).getAutowireCapableBeanFactory();
        organDonorDeceasedCleanupServlet.init(mockServletConfig);

        // Arrange OrganDonorDAO mock - getPersonIdBatch
        Mockito.doReturn(asList("1105010101", "1105020202", "1105030303", "1105040404","1105050505", "1105060606", "1105070707", "1105080808")).when(organDonorDAO).getPersonIdBatch("1105");
        Mockito.doReturn(asList("2310010101", "2310020202", "2310030303", "2310040404")).when(organDonorDAO).getPersonIdBatch("2310");
        Mockito.doReturn(asList("3112010101", "3112020202", "3112030303", "3112040404")).when(organDonorDAO).getPersonIdBatch("3112");


        // Arrange PersonInformation mock - getDeceased
        Mockito.doReturn(asList("1105020202", "1105040404")).when(personInformation).getDeceased(
                Mockito.argThat(l -> l.stream().allMatch(s -> s.startsWith("1105"))),
                Mockito.argThat(i -> !i.isAfter(dateTimeProvider.now().minus(cleanupProperties.getDeceasedPeriod()).toInstant())));

        Mockito.doReturn(asList("2310010101", "2310020202", "2310030303", "2310040404")).when(personInformation).getDeceased(
                Mockito.argThat(l -> l.stream().allMatch(s -> s.startsWith("2310"))),
                Mockito.argThat(i -> !i.isAfter(dateTimeProvider.now().minus(cleanupProperties.getDeceasedPeriod()).toInstant())));

        Mockito.doReturn(asList()).when(personInformation).getDeceased(
                Mockito.argThat(l -> l.stream().allMatch(s -> s.startsWith("3112"))),
                Mockito.argThat(i -> !i.isAfter(dateTimeProvider.now().minus(cleanupProperties.getDeceasedPeriod()).toInstant())));

        Mockito.doThrow(new RuntimeException("Der skal ikke findes personer med andet prefix end 1105, 2310 eller 3112")).when(personInformation).getDeceased(
                Mockito.argThat(l -> l.stream().anyMatch(s -> !s.startsWith("1105") && !s.startsWith("2310") && !s.startsWith("3112"))),
                Mockito.any());

        // Arrange OrganDonorDAO mock - deleteLogEntriesByPersonIdBatch
        Mockito.doReturn(2).
                when(organDonorDAO).
                deleteEntriesByPersonIdBatch(Mockito.argThat(l -> l.stream().allMatch(s -> s.startsWith("1105"))));

        Mockito.doReturn(4).
                when(organDonorDAO).
                deleteEntriesByPersonIdBatch(Mockito.argThat(l -> l.stream().allMatch(s -> s.startsWith("2310"))));

        Mockito.doThrow(new RuntimeException("Der skal ikke slettes nogle personer med andet prefix end 1105 og 2310")).
                when(organDonorDAO).
                deleteEntriesByPersonIdBatch(Mockito.argThat(l -> l.stream().anyMatch(s -> !s.startsWith("1105") && !s.startsWith("2310"))));
    }

    private static List<String> asList(String... elements) {
        return Arrays.asList(elements);
    }

    @Test
    public void testSletningFejlerHvisUventetPersonPrefixTilfoejes() throws Exception{
        // Arrange
        Mockito.doReturn(asList("0101010101")).when(organDonorDAO).getPersonIdBatch("0101");
        // Act
        organDonorDeceasedCleanupServlet.doGet(mockHttpServletRequest, mockHttpServletResponse);
        // Assert
        Mockito.verify(organDonorDAO, Mockito.atMost(366)).getPersonIdBatch(Mockito.anyString());
        Mockito.verify(personInformation, Mockito.atMost(5)).getDeceased(Mockito.anyList(), Mockito.any());
        Mockito.verify(organDonorDAO, Mockito.atMost(3)).deleteEntriesByPersonIdBatch(Mockito.anyList());

        Assert.assertEquals(500, mockHttpServletResponse.getStatus());
        Assert.assertEquals("Internal Error\r\n", mockHttpServletResponse.getContentAsString());
    }

    @Test
    public void testServletAltErGodt() throws Exception{
        // Act
        organDonorDeceasedCleanupServlet.doGet(mockHttpServletRequest, mockHttpServletResponse);
        // Assert
        Mockito.verify(organDonorDAO, Mockito.times(366)).getPersonIdBatch(Mockito.anyString());
        Mockito.verify(personInformation, Mockito.times(3)).getDeceased(Mockito.anyList(), Mockito.any());
        Mockito.verify(organDonorDAO, Mockito.times(2)).deleteEntriesByPersonIdBatch(Mockito.anyList());

        Assert.assertEquals(200, mockHttpServletResponse.getStatus());
        Assert.assertEquals("OK\r\n", mockHttpServletResponse.getContentAsString());
    }

    @Test
    public void testServletAltErGodtMenBatchSizesErMindre() throws Exception{
        //Arrange
        Mockito.doReturn(5).when(cleanupProperties).getPersoninformationDeceasedBatchSize();
        Mockito.doReturn(5).when(cleanupProperties).getDeletionDeceasedBatchSize();

        // Act
        organDonorDeceasedCleanupServlet.doGet(mockHttpServletRequest, mockHttpServletResponse);
        // Assert
        Mockito.verify(organDonorDAO, Mockito.times(366)).getPersonIdBatch(Mockito.anyString());
        Mockito.verify(personInformation, Mockito.times(4)).getDeceased(Mockito.anyList(), Mockito.any());
        Mockito.verify(organDonorDAO, Mockito.times(3)).deleteEntriesByPersonIdBatch(Mockito.anyList());

        Assert.assertEquals(200, mockHttpServletResponse.getStatus());
        Assert.assertEquals("OK\r\n", mockHttpServletResponse.getContentAsString());
    }
}
